/*-
 * Copyright (C) @BABOLO  2008 Apr 14
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2008 Apr 14\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: db_0.c,v 1.8 2008/05/13 11:32:01 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <mife.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

#define INDX(R, C) [R * res->ncol + C]

typedef struct pgoblin_db_0_data {
    u_char  **cell;
    size_t   *llen;
    u_int32_t flags;
#   define    PGOBLIN_DB_0FREEC 0x0001
#   define    PGOBLIN_DB_0FREEA 0x0002
    int       nrow;
    int       ncol;
} pgoblin_db_0_data;

static void *
pgoblin_no() {
    warnx("Unimplemented in 0");
    return(NULL);
}

void *
pgoblin_db_0_ctl(pgoblin_exenv *exenv, pgoblin_r *r) {
    pgoblin_db_0_data *db0;
    int c, i;

    if  (!(db0 = malloc(sizeof(pgoblin_db_0_data)))) {
        errno = ENOMEM;
        goto out;
    }
    bzero(db0, sizeof(pgoblin_db_0_data));
    db0->flags = exenv->options->flags & BLIN_VERMASK;
    db0->nrow = 1;
    for (c = 0; ; c++) {
        i = c;
        if  (c && exenv->pgm->code[exenv->ppoint + c].cmd != PGOBLIN_cont) break;
        if  (!exenv->pgm->code[exenv->ppoint + c].e.lit) break;
    }
    db0->ncol = i;
    if  (i) {
        if  (!(db0->cell = malloc((i + 1) * sizeof(char**)))) {
            free(db0);
            errno = ENOMEM;
            goto out;
        }
        for (c = 0; c < i; c++) db0->cell[c] = exenv->pgm->code[exenv->ppoint + c].e.lit;
        db0->cell[i] = NULL;
    }
out:
    return(db0);
}

static int
pgoblin_shurecon(pgoblin_conn *db) {
    /** R_CONN + **/
    ;;  db->conn = NULL;
    /** R_CONN - **/
    return(EX_OK);
}

static int
pgoblin_query(pgoblin_conn *db, void **res, const char *query, u_char cond) {
    *res = NULL;
    return(EX_OK);
}

static int
pgoblin_ntuples(pgoblin_db_0_data *res) {
    return(res->nrow);
}

static int
pgoblin_nfields(pgoblin_db_0_data *res) {
    return(res->ncol);
}

static int
pgoblin_isnull(pgoblin_db_0_data *res, int row, int col) {
    int e;
    if  (row >= 0 && row < res->nrow && col >= 0 && col < res->ncol) {
        e = !res->cell INDX(row, col);
    } else {
        ifBLIN_QV2(res->flags) {
            warnx( "db 0: row %d %s in [0..%d], col %d %s in [0..%d]"
                 , row, (row >= 0 && row < res->nrow) ? "" : "NOT", res->nrow
                 , col, (col >= 0 && col < res->ncol) ? "" : "NOT", res->ncol
                 );
        }
        e = -1;
    }
    return(e);
}

static int
pgoblin_length(pgoblin_db_0_data *res, int row, int col) {
    int e;
    switch (pgoblin_isnull(res, row, col)) {
    case  1: e = 0;
             break;
    case  0: if  (res->llen) e = res->llen INDX(row, col); else e = strlen(res->cell INDX(row, col));
             break;
    default: e = -1;
             break;
    }
    return(e);
}

static char *
pgoblin_value(pgoblin_db_0_data *res, int row, int col) {
    char *c = NULL;
    if  (!pgoblin_isnull(res, row, col)) c = res->cell INDX(row, col);
    return(c);
}

static void
pgoblin_frees(pgoblin_db_0_data **res) {
    int i;
    if  (res && *res) {
        if  ((*res)->llen) free((*res)->llen);
        if  ((*res)->cell) {
            if  ((*res)->flags & PGOBLIN_DB_0FREEC) {
                for (i = 0; i < (*res)->nrow * (*res)->ncol; i++) free((*res)->cell[i]);
            }
            if  ((*res)->flags & PGOBLIN_DB_0FREEA) free(*(*res)->cell);
            free((*res)->cell);
        }
        free(*res);
        *res = NULL;
}   }

static void *
pgoblin_none(void *v) {
    return(NULL);
}

static int
pgoblin_tarry(void *conn, int timeout) {
    return(0);
}

pgoblin_dbases pgoblin_db_0 =
{ "3pGoblin-" VERS
, "0\0" VERS
,                                      pgoblin_shurecon                            /* shurecon  */
, (int  (*)(pgoblin_conn*, void**, const char*, u_char))pgoblin_query              /* exec      */
, (int  (*)(void*))                    pgoblin_ntuples                             /* ntuples   */
, (int  (*)(void*))                    pgoblin_nfields                             /* nfields   */
, (int  (*)(void*, int, int))          pgoblin_isnull                              /* getisnull */
, (int  (*)(void*, int, int))          pgoblin_length                              /* getlength */
, (char*(*)(void*, int, int))          pgoblin_value                               /* getvalue  */
, (int  (*)(pgoblin_conn*, void**, int, u_char))pgoblin_no                         /* getstream */
, (int  (*)(pgoblin_conn*, void**))    pgoblin_no                                  /* getcopy   */
, (int  (*)(pgoblin_conn*, const void*, int))pgoblin_no                            /* putcopy   */
, (int  (*)(pgoblin_conn*, u_char))    pgoblin_no                                  /* endstream */
, (int  (*)(void*, int))               pgoblin_tarry                               /* tarry     */
, (char*(*)(void*))                    pgoblin_none                                /* notifies  */
, (void (*)(void**))                   pgoblin_none                                /* freecopy  */
, (void (*)(char**))                   pgoblin_none                                /* freenote  */
, (void (*)(void**))                   pgoblin_frees                               /* clear     */
, (void (*)(void**))                   pgoblin_none                                /* finish    */
};
